# TODO: make this into its own package?
# DEPRECIATED
# getScreenResolution (xdpyinfo | grep dimensions)
# verifyInput & collectInputVector: use 'menu' instead.

#' See if an argument is na, null, or length 0 (syntactic sugar).
#' @param x object to check
#' @return boolean, TRUE if it is na, null OR length 0.
#' @note syntactic sugar for \code{is.null(x) || is.na(x) || length(x)==0}.
#' Note that the empty string ("") is NOT "empty" in this sense.
#' @examples
#' is.empty(NA)          # true
#' is.empty(NULL)        # true
#' is.empty(character()) # true, length 0
#' is.empty("")          # false, this has length 1.
#' @export
is.empty <- function(x)
{
    return( is.null(x) || is.na(x) || length(x)==0 )
}

#' gets the last character of a string
#' @param str input string
#' @return the last character of that string.
#' @export
getLastChar <- function( str )
{
    return( substr(str, nchar(str), nchar(str)) )
}

# ******************************** MATHS ******************************** #
#' make 2D rotation matrix anticlockwise by theta
#' @param th amount to rotate by (radians)
#' @return the matrix:
#' \deqn{\left( \begin{array}{cc}
#'      \cos(th) & -\sin(th)\\
#'      \sin(th) & \cos(th)
#'      \end{array}\right)}
#' @note  \code{make2Drot(th) * t(x,y)} gives (x,y) rotated by angle theta
#' anticlockwise.
#' @export
make2Drot <- function(th) {
    return( matrix(c(cos(th),-sin(th),
                     sin(th),cos(th)), ncol=2,  byrow=TRUE) )
} # make2Drot

#' return euclidean distance between successive rows in a matrix.
#' @param m    An nxm matrix where there are n points in \eqn{R^m}, and we
#'             want to find the distance between successive rows (points).
#'             i.e., one point per row.
#' @param circular boolean, whether to also calculate the distance between
#' the last and first point.
#' @param cols If this is TRUE, the matrix is mxn instead (each point is a
#'             column)
#' @return a vector of length (n-1) containing the distance between
#'         successive points in "m".
#' @examples
#' # make a matrix of 4 points in 2D.
#' m <- array(runif(4*2),dim=c(4,2))
#' euclidDist(m)            # dist between rows
#' euclidDist(m,circular=TRUE) # dist between rows including 4 to 1.
#' euclidDist(m,cols=TRUE)     # dist between cols
#' @export
euclidDist <- function(m,cols=FALSE, circular=FALSE) {
    if ( cols )
        m <- t(m)
    idx <- 1:nrow(m)
    if ( circular )
        idx <- c(idx,1)
    return(sqrt(rowSums(diff(as.matrix(m[idx,]))^2)))
}

#' sort vertices in a rectangle in anticlockwise fashion.
#' @param pts           the input points, nx2, colnames 'X' and 'Y'
#' @param anticlockwise boolean, whether to return points in anticlockwise 
#'                      or clockwise fashion (default TRUE)
#' @param verbose       default FALSE, whether to print results
#' @param shift         default FALSE, whether to return results relative 
#'                      to origin
#' @param preserveStart default FALSE, whether to start at the first 
#'                      input coordinate and proceed anticlockwise/
#'                      clockwise from there, or to start from the
#'                      lower-left corner ('origin')
#' @param index.return  default FALSE, whether to return the indices into
#'                      the original "pts" that reorder as specified.
#' @return the pts structure with columns 'X','Y', and rows sorted in
#'         anticlockwise or clockwise order.
#'         If "index.return" is TRUE, returns a list:
#'         \describe{
#'             \item{x}{the points array}
#'             \item{ix}{indices into "pts" that reorder it into output$x}
#'         }
#'         Additionally, the row-names of the output are labelled
#'          "bottom-left", "bottom-right", "top-left", and "top-right",
#'         based on an X-Y coordinate system, 0 being in the bottom left,
#'          increasing moving left-to-right and bottom-to-top.
#' @export
sortPointsRectangle <- function( pts, anticlockwise=TRUE, verbose=FALSE, shift=FALSE, preserveStart=FALSE, index.return=FALSE )
{
    coords <- pts[(idxpts <- order(pts[,"Y"],pts[,"X"])),c("Y","X")] # ordered by Y coordinate (asc) and then X if a tie (asc)
    originI <- 2
    uvL0I   <- 1
    if ( coords[1,"X"]<coords[2,"X"] ) { # then it's coords[1,] that's the origin
        originI <- 1
        uvL0I   <- 2
    }

    uv0LI <- 3
    uvLLI <- 4
    if ( coords[4,"X"]<coords[3,"X"] ) {
        uv0LI <- 4
        uvLLI <- 3
    }
    if ( verbose ) {
	    cat("coords: \n")
	    print(coords)
    	cat("origin       : ", coords[originI,c("X","Y")],"\n")
    	cat("bottom right : ", coords[  uvL0I,c("X","Y")],"\n")
    	cat("top left     : ", coords[  uv0LI,c("X","Y")],"\n")
    	cat("top right    : ", coords[  uvLLI,c("X","Y")],"\n")
    }
    
    idx <- c(originI,uvL0I,uvLLI,uv0LI)
    rn  <- c('bottom-left','bottom-right','top-right','top-left')
    
    if ( preserveStart ) {
        sI <- which( coords[idx,"X"]==pts[1,"X"] & coords[idx,"Y"]==pts[1,"Y"] )
        idx <- idx[ ((sI-1) + 0:3)%%4 + 1 ]
    }

    if ( !anticlockwise ) idx <- idx[c(1,4,3,2)]

    out <- coords[idx,]
    rownames(out) <- rn
    if ( shift ) 
        out[,c("X","Y")] <- t( t(out[,c("X","Y")]) - coords[idx[1],c("X","Y")] )

    out <- out[,c('X','Y')]

    # NOTE: idx is an index in 'coords'
    # ipts is an index in 'pts'
    if ( index.return )
        return(list(x=out,ix=idxpts[idx]))
    else
       return(out) 

} # sortPointsRectangle

#' converts GPS coordinates to UTM coordinates
#' @param gps a nx2 matrix with LON then LAT
#' @param using one of 'proj' or 'rgdal': use proj executable (must be in 
#'              $PATH) or rgdal package
#' @param inverse whether to invert (ie UTM2GPS), default FALSE 
#'               (i.e. GPS -> UTM)
#' @param zone the zone for conversion. 
#'        If doing GPS to UTM conversion, this will be calculated using 
#'        the mean longitude.
#' @param hemisphere hemisphere for conversion.
#'        If doing GPS to UTM conversion, this will be calculated using 
#'        the latitude.
#' @return a nx2 matrix with utm EASTING then utm NORTHING (x,y)
#' @details
#' You need either package 'rgdal' [need libgdal1-dev in ubuntu and libproj-dev] OR 'proj' in your path [proj-bin in ubuntu].
#'
#' If doing conversion from UTM to GPS, it is *very important* to 
#' know the zone and the hemisphere, as an easting/northing pair can be
#' valid in both hemispheres, and/or multiple zones.
#' 
#' If doing conversion from GPS to UTM, it's not as improtant to know the
#' zone/hemisphere, as the hemisphere can be told from the latitude.
#' 
#' The zone can be calculated as (wikipedia):
#' 
#' The UTM system divides the surface of the Earth ...
#'  into 60 zones, each of 6 degrees of longitude in width.
#' Zone 1 covers longitude 180 to 174 degrees W; zone numbering increases
#' eastward to zone 60 that covers longitude 174 to 180 E.
#'
#' @family coordinates
#' @export
# @importFrom rgdal project
GPS2UTM <- function( gps, using=c('proj','rgdal'), inverse=FALSE, zone=56, hemisphere=c('south','north') )
{
    # work out whether to use rgdal or not
    .USING_RGDAL <- NA
    PROJ <- getProgram(c('proj','proj.exe'))
    if ( missing(using) ) { # automagically determine what to use
        .USING_RGDAL <- is.null(PROJ)
    } else {
        .USING_RGDAL <- match.arg(tolower(using),c('proj','rgdal'))=='rgdal'
    }

    # check for the required components.
    if ( .USING_RGDAL && !require(rgdal) ) 
        stop("You need to install the rgdal package.\n(On Ubuntu, this requires libgdal1-dev and libproj-dev).\nExiting\n")
    if ( !.USING_RGDAL && is.null(PROJ) )
        stop('You need to install proj: sudo (yum|apt-get|etc) proj (on ubuntu, \'proj-bin\'). If it\'s already installed, then you need to add it to your $PATH.\nAlternatively, install package rgdal to R: install.packages("rgdal")\nExiting.\n')
  
    # NEW: automagically determine zone?
    # The UTM system divides the surface of Earth between 80°S and 84°N latitude into 60 zones, 
    # each 6 degrees of longitude in width. 
    # Zone 1 covers longitude 180 to 174 degrees W; 
    # zone numbering increases eastward to zone 60 that covers longitude 174 to 180 East.
    #
    # '%%60' in case longitudes > 180 or < -180
    # will use mean longitude.
    if (missing(zone) && !inverse) {
        zone <- (floor((mean(gps[,1]) + 180)/6) + 1) %% 60
    } 
    # NOTE: for *backwards* transformations you have to pass in the zone :/

    # NEW: determine hemisphere in forward direction...check latitude.
    if ( !inverse ) {
        HEMISPHERE_ARG <- ifelse( gps[1,2]>=0,'','+south' )
    } else {
        hemisphere <- match.arg(hemisphere,c('south','north'))
        HEMISPHERE_ARG <- ifelse(hemisphere=='south','+south','')
    }
       
    # specify projectiona arguments.
    PROJ_ARGS <- sprintf("+proj=utm +datum=WGS84 +zone=%i %s -f '%%f'",zone,HEMISPHERE_ARG) # arguments to pass in to proj. '%' escaped for sprintf.

    # conert
    if ( .USING_RGDAL ) {
        utms <- project( as.matrix(gps), PROJ_ARGS,inv=inverse )
    } else { # use command line easier.
        if ( inverse )
            PROJ <- paste(PROJ,'-I')

        cmd <- sprintf("%s %s <<EOF
%s
EOF",PROJ,PROJ_ARGS,paste(gps[,1],gps[,2],collapse='\n'))
    # separate out. (1-based) indices 1,3,.. are x/easting, 2,4,... are y/northing
        utmstxt <- system(cmd,intern=TRUE)  # x\ty  or easting\tnorthing
        utms <- as.numeric(unlist(strsplit(utmstxt,'\t')))
        utms <- matrix(utms,ncol=2,byrow=TRUE)
    }
    if ( inverse )
        colnames(utms)<-c("lon","lat")
    else 
        colnames(utms)<-c("X","Y")

    return(utms)
} # GPS2UTM

#' converts UTM coordinates to GPS coordinates
#' @param utm a nx2 matrix with X then Y
#' @param ... params for UTM2GPS
#' @return a nx2 matrix with gps LONGITUDE then gps LATITUDE
#' @details
#' You need either package 'rgdal' [need libgdal1-dev in ubuntu and libproj-dev] OR 'proj' in your path [proj-bin in ubuntu]
#' @family coordinates
#' @export
UTM2GPS <- function( utm, ... )
    return(GPS2UTM( utm,inv=TRUE, ...))

#' formats an nx2 array with `lon,lat` to "\%.10fE, \%.10fS".
#'
#' @inheritParams GPS2UTM
#' @param verbose whether to print the output to screen.
#' @return a single string containing the lon,lat in format 
#'         "\%.10fE, \%.10fS", each coordinate separted by a new line.
#' @family coordinates
#' @export
formatGPS <- function(gps,verbose=TRUE) {
    str = paste(apply(gps,1,function(x)sprintf("%.10fE, %.10fS", x[1],-x[2])),collapse='\n')
    if (verbose)
        cat(str)
    return(str)
}

#' asks a question and waits for valid input.
#' 
#' This is a bit like \code{\link[utils]{menu}} except it doesn't display
#' the choices as \code{1: xx 2: xx} etc. Suitable for yes/no questions
#' where 'menu' is a bit too much.
#'
#' @param question the question to ask the user
#' @param ... the list of valid inputs to chose from, of the form
#' \code{input=action}. This is the same sort of input that should be fed
#' into \code{\link[base]{switch}}. All arguments must be names, except
#' possibly one to be used as the default value (which must occur last).
#' See details.
#' @param extraArgs a *list* of extra arguments to pass to the function
#' that gets chosen from \code{...}. The format of the list is as for
#' \code{\link[base]{do.call}}.
#' @return the value of one of the elements of \code{...}.
#' 
#' @details
#' The answer to \code{question} is collected as a string using
#' \code{\link[base]{readline}}. 
#' 
#' This is then fed (in string form) to \code{\link[base]{switch}} with the
#' \code{...} arguments: \code{switch( readline(question), ... )}.
#'
#' As such, the default argument (if any) must go last.
#'
#' The argument that is selected is:
#' \itemize{
#'     \item evaluated via \code{do.call(arg,extraArgs)}, if it is a
#'     function.
#'     \item simply returned, otherwise.
#' }
#' The output of this is returned as the output of the function.
#'
#' @examples
#' \dontrun{
#' # example of setting a parameter via a yes/no question:
#' p <- 1 
#' p <- verifyInput( sprintf('current value of p is %i, keep? [Y/N] ',p),
#'                   Y=,y=p,                    # return p if they keep
#'                   N=,n= function() readline('Enter new value for p: '))
#' # If the user says 'N' they enter a new value for p and this is
#' # returned.
#' }
#' @export
verifyInput <- function( question, ..., extraArgs=list() )
{
    while (TRUE) {
        fun <- switch(readline(question),...)
        if ( !is.null(fun) )
            if ( is.function(fun) )
                return( do.call(fun, extraArgs) ) 
            else
                return(fun)
    }
}


# ************** DEPRECIATED ******************* #
####################################
#verifyInput <- function( question, cases, case.sensitive=FALSE,
#                        nchars=-1,... ) 
#
####################################
#{
#    viFinished <- FALSE
#    # convert any numbers to strins
#    names(cases) <- as.character(names(cases))
#    if ( !case.sensitive ) names(cases) <- tolower(names(cases))
#
#    while ( !viFinished ) { 
#        dmy <- readline( question )
#        fun <- switch(
#        if ( nchars > 0 ) { dmy <- substr(dmy,1,nchars) }
#        if ( !case.sensitive ) { dmy <- toupper(dmy) }
#        for ( response in names(cases) ) {
#            if ( dmy == ifelse(case.sensitive,response,toupper(response)) ) {
#                viFinished <- TRUE
#                eval(parse(text=cases[[response]]),parent.frame()) 
#                break
#                # eval(parse(text=cases[[response]])) 
#            } else if ( response == "default" ) {
#                viFinished <- TRUE
#                eval(parse(text=cases[[response]],parent.frame()))
#                break
#            }
#        } # for response in names(cases)
#    } # while (!viFinished)
#    return(dmy)
#} # verifyInput
#
####################################
#collectInputVector <- function(question, cases, vectorType="character", case.sensitive=FALSE ) 
####################################
## Will ask [question] until a 'finished <- TRUE' is evaluated. Amalgamates responses in a vector.
## Input is compared against [cases], a list (e.g. cases=list('Y'='doSomething','N'='doSomethingElse', 'default'='somethingElseAgain'))
## The names of the list are compared to the input and if they match the value is eval'd in the FUNCTION (not parent) ENVIRONMENT.
## In order for this function to return you need a 'finished <- TRUE' in at least one of your cases 
##  (once this is encountered the function will return all the responses thus far)
## You can have a case 'default' in cases which MUST BE LAST.
#{
#    v        <- vector(mode=vectorType) 
#    finished <- FALSE
#    while (!finished) {
#        dmy <- readline(question)
#        v[length(v)+1] <- as( dmy, vectorType )
#        if ( !case.sensitive ) { dmy <- toupper(dmy) }
#        
#        # loop through all other possibilities first
#        for ( response in names(cases) ) {
#            if ( dmy == response ) {
#                eval(parse(text=cases[[response]])) 
#                break
#            } else if ( response == "default" ) {
#                eval(parse(text=cases[[response]])) 
#                break
#            }
#        }
#    } # while (!finished) 
#
#    return(v)
#} # collectInput
